module Transpiler
using CUDA
using ..ExpressionProcessing
using ..Utils

# https://docs.nvidia.com/cuda/cuda-c-programming-guide/#features-and-technical-specifications

const BYTES = sizeof(Float32)
const Operand = Union{Float32, String} # Operand is either fixed value or register
cache = Dict{Expr, CuFunction}() # needed if multiple runs with the same expr but different parameters are performed

function evaluate(expressions::Vector{Expr}, variables::Matrix{Float32}, parameters::Vector{Vector{Float32}})
	varRows = size(variables, 1)
	variableCols = size(variables, 2)
	kernels = Vector{CuFunction}(undef, length(expressions))
	
	# Test this parallel version again when doing performance tests. With the simple "functionality" tests this took 0.03 seconds while sequential took "0.00009" seconds
	# Threads.@threads for i in eachindex(expressions)
	#   TODO: Use cache
	# 	kernel = transpile(expressions[i], varRows, Utils.get_max_inner_length(parameters))

	# 	linker = CuLink()
	# 	add_data!(linker, "ExpressionProcessing", kernel)

	# 	image = complete(linker)
	
	# 	mod = CuModule(image)
	# 	kernels[i] = CuFunction(mod, "ExpressionProcessing")
	# end

	for i in eachindex(expressions)
		if haskey(cache, expressions[i])
			kernels[i] = cache[expressions[i]]
			continue
		end

		formattedExpr = ExpressionProcessing.expr_to_postfix(expressions[i])
		kernel = transpile(formattedExpr, varRows, Utils.get_max_inner_length(parameters), variableCols, i)
		
		println(kernel)

		linker = CuLink()
		add_data!(linker, "ExpressionProcessing", kernel)
		
		image = complete(linker)
		
		mod = CuModule(image)
		kernels[i] = CuFunction(mod, "ExpressionProcessing")
		cache[expressions[i]] = kernels[i]
	end

	cudaVars = CuArray(variables) # maybe put in shared memory (see runtests.jl for more info)
	cudaParams = Utils.create_cuda_array(parameters, NaN32) # maybe make constant (see runtests.jl for more info)

	# each expression has nr. of variable sets (nr. of columns of the variables) results and there are n expressions
	cudaResults = CuArray{Float32}(undef, variableCols * length(expressions))
	# cudaResults = CUDA.zeros(variableCols * length(expressions))
	# ptr = CuPtr{Float32}(C_NULL)
	# CUDA.cuMemAlloc(ptr, sizeof(Float32) * 10)
	# values = Float32[1.0, 2.0, 3.0, 4.0, 5.0, 6.0, 7.0, 8.0, 9.0, 10.0]
	# CUDA.cuMemcpyHtoD(ptr, values, sizeof(Float32) * 10)
	# copyto!(cudaResults, z)

	# execute each kernel (also try doing this with Threads.@threads. Since we can have multiple grids, this might improve performance)
	for i in eachindex(kernels)
		config = launch_configuration(kernels[i])
		threads = min(variableCols, config.threads)
		blocks = cld(variableCols, threads)

		# cudacall(kernels[i], (CuPtr{Float32},CuPtr{Float32},CuPtr{Float32}), cudaVars, cudaParams, cudaResults; threads=threads, blocks=blocks)
		cudacall(kernels[i], (CuPtr{Float32},), cudaResults; threads=threads, blocks=blocks)
		# launch(kernels[i], cudaVars, cudaParams, cudaResults; threads=threads, blocks=blocks)
		# launch(kernels[i], cudaResults; threads=threads, blocks=blocks)
	end
	
	println(Array(cudaResults))
	# cudaResults = nothing
	# CUDA.cuMemFree(ptr)
end

# To increase performance, it would probably be best for all helper functions to return their IO Buffer and not a string
# seekstart(buf1); write(buf2, buf1)
"
- param ```varSetSize```: The size of a variable set. Equal to number of rows of variable matrix (in a column major matrix)
- param ```paramSetSize```: The size of the longest parameter set. As it has to be stored in a column major matrix, the nr of rows is dependent oon the longest parameter set
- param ```expressionIndex```: The 1-based index of the expression
"
function transpile(expression::ExpressionProcessing.PostfixType, varSetSize::Integer, paramSetSize::Integer, 
				   nrOfVariableSets::Integer, expressionIndex::Integer)::String
	exitJumpLocationMarker = "\$L__BB0_2"
	ptxBuffer = IOBuffer()

	# TODO: Temp fix. Make these types and create new instances for every call to this function. Otherwise we cannot parallelise the transpilation
	reset_registers()
	reset_symtable()

	# TODO: Suboptimal solution
	signature, paramLoading = get_kernel_signature("ExpressionProcessing", [Float32]) # Vars, Params, Results
	guardClause, threadIdReg = get_guard_clause(exitJumpLocationMarker, nrOfVariableSets)

	println(ptxBuffer, get_cuda_header())
	println(ptxBuffer, signature)
	println(ptxBuffer, "{")


	calc_code = generate_calculation_code(expression, "%parameter0", varSetSize, "%parameter1", paramSetSize, "%parameter0", 
										  threadIdReg, expressionIndex, nrOfVariableSets)
	println(ptxBuffer, get_register_definitions())
	println(ptxBuffer, paramLoading)
	println(ptxBuffer, guardClause)
	println(ptxBuffer, calc_code)

	# exit jump location
	print(ptxBuffer, exitJumpLocationMarker); println(ptxBuffer, ": ret;")
	println(ptxBuffer, "}")

	generatedCode = String(take!(ptxBuffer))
	return generatedCode
end

# TODO: Make version, target and address_size configurable; also see what address_size means exactly
function get_cuda_header()::String
	return "
.version 8.5
.target sm_61
.address_size 64
"
end

"
param ```parameters```: [1] = nr of var sets; [2] = variables; [3] = parameters; [4] = result
"
function get_kernel_signature(kernelName::String, parameters::Vector{DataType})::Tuple{String, String}
	signatureBuffer = IOBuffer()
	paramLoadingBuffer = IOBuffer()
	print(signatureBuffer, ".visible .entry ")
	print(signatureBuffer, kernelName)
	println(signatureBuffer, "(")
	
	for i in eachindex(parameters)
		print(signatureBuffer, "  .param .u64", " ", "param_", i)

		parametersLocation = get_next_free_register("i")
		println(paramLoadingBuffer, "ld.param.u64   $parametersLocation, [param_$i];")
		println(paramLoadingBuffer, "cvta.to.global.u64   $(get_next_free_register("parameter")), $parametersLocation;")
		if i != lastindex(parameters)
			println(signatureBuffer, ",")
		end
	end

	print(signatureBuffer, ")")
	return (String(take!(signatureBuffer)), String(take!(paramLoadingBuffer)))
end

"
Constructs the PTX code used for handling the case where too many threads are started.

- param ```nrOfVarSetsRegister```: The register which holds the total amount of variable sets for the kernel
"
function get_guard_clause(exitJumpLocation::String, nrOfVarSets::Integer)::Tuple{String, String}
	guardBuffer = IOBuffer()

	threadIds = get_next_free_register("r")
	threadsPerCTA = get_next_free_register("r")
	currentThreadId = get_next_free_register("r")

	# load data into above defined registers
	println(guardBuffer, "mov.u32    $threadIds, %ntid.x;")
	println(guardBuffer, "mov.u32    $threadsPerCTA, %ctaid.x;")
	println(guardBuffer, "mov.u32    $currentThreadId, %tid.x;")

	globalThreadId = get_next_free_register("r") # basically the index of the thread in the variable set
	breakCondition = get_next_free_register("p")
	println(guardBuffer, "mad.lo.s32     $globalThreadId, $threadIds, $threadsPerCTA, $currentThreadId;")
	println(guardBuffer, "setp.gt.s32    $breakCondition, $globalThreadId, $nrOfVarSets;") # guard clause = index > nrOfVariableSets

	# branch to end if breakCondition is true
	print(guardBuffer, "@$breakCondition bra    $exitJumpLocation;")

	return (String(take!(guardBuffer)), globalThreadId)
end

"
- param ```parametersSetSize```: Size of the largest parameter set
"
function generate_calculation_code(expression::ExpressionProcessing.PostfixType, variablesLocation::String, variablesSetSize::Integer, 
								   parametersLocation::String, parametersSetSize::Integer, resultsLocation::String, 
								   threadIdReg::String, expressionIndex::Integer, nrOfVarSets::Integer)::String

	# return "st.global.f32  [$resultsLocation], 10.0;"

	codeBuffer = IOBuffer()
	operands = Vector{Operand}()

	for i in eachindex(expression)
		token = expression[i]

		if token.Type == FLOAT32
			push!(operands, reinterpret(Float32, token.Value))
		elseif token.Type == OPERATOR
			operator = reinterpret(Operator, token.Value)

			right = nothing
			if is_binary_operator(operator)
				right = pop!(operands)
				left = pop!(operands)
			else
				left = pop!(operands)
			end
			operation, resultRegister = get_operation(operator, left, right)
			
			println(codeBuffer, operation)
			push!(operands, resultRegister)
		elseif token.Type == INDEX
			if token.Value > 0 # varaibles
				var, first_access = get_register_for_name("x$(token.Value)")
				if first_access
					println(codeBuffer, load_into_register(var, variablesLocation, token.Value, threadIdReg, variablesSetSize))
				end
				push!(operands, var)
			else
				absVal = abs(token.Value)
				param, first_access = get_register_for_name("p$absVal")
				if first_access
					println(codeBuffer, load_into_register(param, parametersLocation, absVal, threadIdReg, parametersSetSize))
				end
				push!(operands, param)
			end
		end
	end

	# resultIndex = ((expressionIndex - 1) * nrOfVarSets + threadIdReg) * bytes (4 in our case)
	# resultsLocation[resultIndex] = "";
	# tempReg = get_next_free_register("i")
	# println(codeBuffer, "
	# add.u32        $tempReg, $((expressionIndex-1)*nrOfVarSets), $threadIdReg;
	# mul.lo.u32     $tempReg, $tempReg, $BYTES;
	# add.u32        $tempReg, $resultsLocation, $tempReg;
	# st.global.f32  [$tempReg], $(pop!(operands));
	# ")
	println(codeBuffer, "st.global.f32  [$resultsLocation], 10.0;")

	return String(take!(codeBuffer))
end

"
Loads a value from a location into the given register. It is assumed that the location refers to a column-major matrix

- param ```register```: The register where the loaded value will be stored
- param ```loadLocation```: The location from where to load the value
- param ```valueIndex```: 0-based index of the value in the variable set/parameter set
- param ```setIndexReg```: 0-based index of the set. Needed to calculate the actual index from the ```valueIndex```. Is equal to the global threadId
- param ```setSize```: The size of one set. Needed to calculate the actual index from the ```valueIndex```. Total number of elements in the set (length(set))
"
function load_into_register(register::String, loadLocation::String, valueIndex::Integer, setIndexReg::String, setSize::Integer)::String
	# loadLocation + startIndex + valueIndex * bytes (4 in our case)
	# startIndex: setIndex * setSize
	tempReg = get_next_free_register("i")
	return "
	mul.lo.u32 $tempReg, $setIndexReg, $setSize;
	add.u32 $tempReg, $tempReg, $(valueIndex*BYTES);
	add.u32 $tempReg, $loadLocation, $tempReg;
	ld.global.f32 $register, [$tempReg];"
end

function type_to_ptx_type(type::DataType)::String
	if type == Int64
		return ".s64"
	elseif type == Int32
		return ".s32"
	elseif type == Float32
		return ".f32"
	else
		return ".b64"
	end
end

function get_operation(operator::Operator, left::Operand, right::Union{Operand, Nothing} = nothing)::Tuple{String, String}
	resultRegister = get_next_free_register("f")
	resultCode = ""

	if is_binary_operator(operator) && isnothing(right)
		throw(ArgumentError("Given operator '$operator' is a binary operator. However only one operator has been given."))
	end

	if operator == ADD
		resultCode = "add.f32    $resultRegister, $left, $right;"
	elseif operator == SUBTRACT
		resultCode = "sub.f32    $resultRegister, $left, $right;"
	elseif operator == MULTIPLY
		resultCode = "mul.f32    $resultRegister, $left, $right;"
	elseif operator == DIVIDE
		resultCode = "div.approx.f32    $resultRegister, $left, $right;"
	elseif operator == POWER
		# x^y == 2^(y*log2(x)) as generated by nvcc for "pow(x, y)"
		resultCode = "
		// x^y:
		lg2.approx.f32   $resultRegister, $left;
		mul.f32          $resultRegister, $right, $resultRegister;
		ex2.approx.f32   $resultRegister, $resultRegister;"
	elseif operator == ABS
		resultCode = "abs.f32    $resultRegister, $left;"
	elseif operator == LOG
		# log(x) == log2(x) * ln(2) as generated by nvcc for "log(x)"
		resultCode = "
		// log(x):
		lg2.approx.f32   $resultRegister, $left;
		mul.f32          $resultRegister, $resultRegister, 0.693147182;"
	elseif operator == EXP
		# e^x == 2^(x/ln(2)) as generated by nvcc for "exp(x)"
		resultCode = "
		// e^x:
		mul.f32          $resultRegister, $left, 1.44269502; 
		ex2.approx.f32   $resultRegister, $resultRegister;"
	elseif operator == SQRT
		resultCode = "sqrt.approx.f32    $resultRegister, $left;"
	else
		throw(ArgumentError("Operator conversion to ptx not implemented for '$operator'"))
	end
	return (resultCode, resultRegister)
end

let registers = Dict() # stores the count of the register already used.
	global get_next_free_register
	global get_register_definitions
	global reset_registers

	function reset_registers()
		registers = Dict()
	end


	# By convention these names correspond to the following types:
	# - p -> pred
	# - f -> float32
	# - r -> 32 bit
	# - var -> float32 (used for variables and params)
	function get_next_free_register(name::String)::String
		if haskey(registers, name)
			registers[name] += 1
		else
			registers[name] = 1
		end

		return string("%", name, registers[name] - 1)
	end

	function get_register_definitions()::String
		registersBuffer = IOBuffer()
	
		for definition in registers
			regType = ""
			if definition.first == "p"
				regType = ".pred"
			elseif definition.first == "f"
				regType = ".f32"
			elseif definition.first == "var"
				regType = ".f32"
			elseif definition.first == "param"
				regType = ".f32"
			elseif definition.first == "r"
				regType = ".b32"
			elseif definition.first == "parameter"
				regType = ".b64"
			elseif definition.first == "i"
				regType = ".b64"
			else
				throw(ArgumentError("Unknown register name used. Name '$(definition.first)' cannot be mapped to a PTX type."))
			end
			println(registersBuffer, ".reg $regType   %$(definition.first)<$(definition.second)>;")
		end
	
		return String(take!(registersBuffer))
	end
end

let symtable = Dict()
	global get_register_for_name
	global reset_symtable

	function reset_symtable()
		symtable = Dict()
	end

	"Returns the register for this variable/parameter and true if it is used for the first time and false otherwise."
	function get_register_for_name(varName::String)
		if haskey(symtable, varName)
			return (symtable[varName], false)
		else
			reg = get_next_free_register("var")
			symtable[varName] = reg
			return (reg, true)
		end
	end
end

end

